use io;

/* Read instance data */
function input() {
    local usage = "Usage: localsolver mdvrp.lsp inFileName=inputFile" 
            + "[solFileName=outputFile] [lsTimeLimit=timeLimit]";
    if (inFileName == nil) throw usage;

    // Read the data and compute matrices
    readInputMdvrp();
    computeDistanceMatrix();
}

/* Declare the optimization model */
function model() {
    // Sequence of customers visited by each truck
    customersSequences[d in 0...nbDepots][k in 0...nbTrucksPerDepot] <- list(nbCustomers);

    // All customers must be visited by exactly one truck
    constraint partition[d in 0...nbDepots][k in 0...nbTrucksPerDepot](customersSequences[d][k]);

    for [d in 0...nbDepots][k in 0...nbTrucksPerDepot] {
        local sequence <- customersSequences[d][k];
        local c <- count(sequence);

        // The quantity needed in each route must not exceed the truck capacity
        routeQuantity <- sum(0...c, i => demands[sequence[i]]);
        constraint routeQuantity <= truckCapacity[d];

        // Distance traveled by truck k of depot d
        routeDistances[d][k] <- sum(1...c, i => distanceMatrix[sequence[i - 1]][sequence[i]])
             + (c > 0 ? (distanceDepot[d][sequence[0]] + distanceDepot[d][sequence[c - 1]]) : 0);

        // We add service Time
        routeServiceTime <- sum(0...c, i => serviceTime[sequence[i]]);
        routeDuration <- routeDistances[d][k] + routeServiceTime;

        // The total distance of this truck should not exceed the duration capacity of the truck
        // (only if we define such a capacity)
        if (routeDurationCapacity[d] > 0) {
            constraint routeDuration <= routeDurationCapacity[d];
        }
    }

    // Total distance traveled
    totalDistance <- sum[d in 0...nbDepots][k in 0...nbTrucksPerDepot](routeDistances[d][k]);

    // Objective: minimize the total distance traveled
    minimize totalDistance;
}

function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

// Write the solution in a file with the following format:
//  - instance, time_limit, total distance
//  - for each depot and for each truck in this depot, the customers visited 
function output() {
    if (solFileName == nil) return;
    local resfile = io.openWrite(solFileName);

    resfile.println("Instance: "+ inFileName + " ; time_limit: " +
            lsTimeLimit + " ; Objective value: " + totalDistance.value);
    for [d in 0...nbDepots] {
        nbTrucksUsed = 0;
        for [k in 0...nbTrucksPerDepot] {
            if (customersSequences[d][k].value.count() > 0) {
                nbTrucksUsed += 1;
            }
        }
        if (nbTrucksUsed > 0){
            n = 1;
            resfile.println("Depot " + (d + 1));
            for [k in 0...nbTrucksPerDepot] {
                if (customersSequences[d][k].value.count() > 0) {
                    resfile.print("Truck " + n + " : ");
                    for [customer in customersSequences[d][k].value] {
                        resfile.print(customer + 1, " ");
                    }
                    n += 1;
                    resfile.println();
                }
            }
            resfile.println();
        }
    }
}

// Input files following "Cordeau"'s format
function readInputMdvrp() {
    local inFile = io.openRead(inFileName);
    inFile.readInt();  // We ignore the first int of the instance

    // Numbers of trucks per depot, customers and depots
    nbTrucksPerDepot = inFile.readInt();
    nbCustomers = inFile.readInt();
    nbDepots = inFile.readInt();

    for [d in 0...nbDepots] {
        routeDurationCapacity[d] = inFile.readInt(); // Time capacity for every type of truck from every depot
        truckCapacity[d] = inFile.readInt(); // Capacity for every type of truck from every depot
    }

    for [n in 0...nbCustomers] {
        inLine = inFile.readln();
        line = inLine.split();

        // Coordinates X and Y, service time and demand for customers
        nodesX[n] = line[1].toDouble();
        nodesY[n] = line[2].toDouble();
        serviceTime[n] = line[3].toInt();
        demands[n] = line[4].toInt();
    }

    for [d in 0...nbDepots] {
        inLine = inFile.readln();
        line = inLine.split();

        // Coordinates X and Y for depots
        depotX[d] = line[1].toDouble();
        depotY[d] = line[2].toDouble();
    }
}

// Compute the distance matrix for customers and for depots/warehouses
function computeDistanceMatrix() {
    for [i in 0...nbCustomers] {
        distanceMatrix[i][i] = 0;
        for [j in (i + 1)...nbCustomers] {
            local localDistance = computeDist(i, j);
            distanceMatrix[j][i] = localDistance;
            distanceMatrix[i][j] = localDistance;
        }
    }

    for [i in 0...nbCustomers][d in 0...nbDepots] {
        local localDistance = computeDistDepot(d, i);
        distanceDepot[d][i] = localDistance;
    }
}

// Compute the distance between two points
function computeDist(i, j) {
    return sqrt(pow(nodesX[i] - nodesX[j], 2) + pow(nodesY[i] - nodesY[j], 2));
}

function computeDistDepot(d, j) {
    return sqrt(pow(depotX[d] - nodesX[j], 2) + pow(depotY[d] - nodesY[j], 2));
}
